/* Nessus Attack Scripting Language
 *
 * Copyright (C) 1999 - 2002 Renaud Deraison
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <includes.h>
#include "strutils.h"
#include "defines.h"
/*
 * The NASL memory manager is used to keep track of the allocated
 * pointers. It allows us to free what has not been freed when
 * we are in nasl_exit().
 *
 * The current implementation is very slow, this should be 
 * corrected in a couple of days
 */

#undef MEMORY_ALERT	/* Useful */
#undef MEMORY_MSG	/* Useless */
#undef SLOW		/* corrects buggy behavior */


/*---------------------------------------------------------------------*/

/*
 * Our memory manager as its own special hash table structure. 
 */

#define HASH_MAX 32769

struct hash {
	void * value;
	struct hash * next;
	};
	
	
static int mkhash(void * ptr)
{
 int h = 0;
 char * p = (char*)&ptr;
 int i;
 
 if( ptr == NULL )
  return 0;
 
 for (i = 0 ; i < sizeof(ptr) ; i ++)
 {
  h = (h * HASH_MAX + p[0]) % HASH_MAX;
 }
 

 if(h < 0)
  return -h;
 else
  return h;
}	

static struct hash * hash_init()
{
 struct hash * h = emalloc(sizeof(*h) * HASH_MAX + 1);

 return h;
}

static void hash_link_destroy(struct hash * h)
{
 if(h == NULL)
  return;
  
 if(h->next != NULL)
  hash_link_destroy(h->next);
 if(h->value != NULL)
 	efree(&h->value);
 efree(&h);
}

static void hash_destroy(struct hash * h)
{
 int i;
 
 for(i=0;i<HASH_MAX;i++)
  {
  hash_link_destroy(h[i].next);
  if(h[i].value != NULL)
 	efree(&(h[i].value));
  }
 efree(&h);
}

static int hash_add(struct hash * h, void * value)
{
 int idx = mkhash(value);
 if(h[idx].value != NULL)
 {
  struct hash * l = emalloc(sizeof(*l));
  struct hash * old;
 
  l->value = value;
  old = h[idx].next;
  h[idx].next = l;
  l->next = old;
 }
 else
 {
  h[idx].value = value;
 }
 return idx;
}





static int _hash_rm(struct hash *h, void *value, int soft)
{
 int idx = mkhash(value);
 struct hash * l = &(h[idx]);
 
 while(l != NULL)
 {
 if(l->value && l->value == value)
 {
  if(soft == 0) 
  	efree(&l->value);
  if(l->next)
  {
   struct hash * o;
   o = l->next;
   l->value = o->value;
   l->next = o->next;
   efree(&o);
  }
  else
   {
   l->value = NULL;
  }
  return 0;
 }
 else l = l->next;
 }
 return -1;
}


static int hash_soft_rm(struct hash *h, void *value)
{
 return _hash_rm(h, value, 1);
}

static int hash_rm(struct hash *h, void *value)
{
 return _hash_rm(h, value, 0);
}

static void * hash_get(struct hash * h, void * value)
{
 int idx = mkhash(value);
 struct hash * l = &(h[idx]);
 while(l != NULL)
 {
  if(l->value && l->value == value)
   return l->value;
  else
   l = l->next;
 }
 return NULL;
}

/* ----------------------------------------------------------------------

	  	     NASL memory management routines
		
   ______________________________________________________________________
   	

   Since the NASL interpreter makes a lot of malloc() and free(),
   and since we do not want any memory leak, I have defined the
   functions nasl_malloc() and nasl_free() which keep track of the
   allocated pointers.
   
 */
 
/*
 * Init the memory manager
 */
struct hash * nasl_init_memory()
{
  return hash_init();
}


/*
 * Allocate memory
 */
void * nasl_malloc(globals, size)
  harglst * globals;
  size_t size;
{
  struct hash * mm = harg_get_ptr(globals, "memory_manager");
  void * buffer = emalloc(size);
  hash_add(mm, buffer);
  return buffer;
}

void * nasl_realloc(globals, buf, size)
 harglst * globals;
 void * buf;
 size_t size;
{
 struct hash * mm = harg_get_ptr(globals, "memory_manager");
 void * buffer = erealloc(buf, size);
 
 hash_soft_rm(mm, buf);
 hash_add(mm, buffer);
 return buffer;
}

/*
 * Duplicate a string (see also nstrdup in
 * strutils.h)
 */
char * nasl_strdup(globals, ptr)
 harglst * globals;
 char * ptr;
{
  size_t size = strlen(ptr);
  struct hash * mm = harg_get_ptr(globals, "memory_manager");
  void * buffer = emalloc(size+1);
  
  hash_add(mm, buffer);
  strncpy(buffer, ptr, size);
  return buffer;
}



/*
 * Free memory
 */
void nasl_free(globals, ptr)
  harglst * globals;
  void * ptr;
{
  struct hash * mm = harg_get_ptr(globals, "memory_manager");
  char * ptr2;
  
  if(ptr == NULL)
   return;
   
  if(!(ptr2 = hash_get(mm, ptr)))
  {
    return;
  }
  
  hash_rm(mm, ptr);
}


void nasl_memory_cleanup(globals)
 harglst * globals;
{
 struct hash * mm = harg_get_ptr(globals, "memory_manager");
 hash_destroy(mm);
}






void nasl_arg_add_value(globals, arglist, name, type, length, value)
 harglst * globals;
 struct arglist * arglist;
 char * name;
 int type;
 int length;
 char * value;
{
 struct arglist * u = arglist;
 while(u->next)u = u->next;

 u->name = nasl_strdup(globals, name);
 if(type & VAR_INT)u->value = (void*)value;
 else u->value = nstrdup(globals, value, length, 0);
 u->type = type;
 u->length = length;
 u->next = nasl_malloc(globals, sizeof(struct arglist));
}


